import os
import sys
import inspect
import asyncio
import warnings

from functools import wraps
from types import FunctionType
from contextlib import _GeneratorContextManager

from apistellar.helper import proxy, get_callargs


def contextmanager(func):
    @wraps(func)
    def helper(*args, **kwds):
        return _AsyncGeneratorContextManager(func, args, kwds)

    return helper


class _AsyncGeneratorContextManager(_GeneratorContextManager):
    def __enter__(self):
        try:
            # 对于同步的方式驱动异步生成器，则什么也不返回，后续在业务层面警告
            if inspect.isasyncgen(self.gen):
                return
            else:
                return self.gen.send(None)
        except StopIteration:
            raise RuntimeError("generator didn't yield") from None

    def __exit__(self, type, value, traceback):
        if type is None:
            try:
                if inspect.isgenerator(self.gen):
                    self.gen.send(None)
                else:
                    return False
            except StopIteration:
                return False
            else:
                raise RuntimeError("generator didn't stop")
        else:
            if value is None:
                # Need to force instantiation so we can reliably
                # tell if we get the same exception back
                value = type()
            try:
                if inspect.isgenerator(self.gen):
                    self.gen.throw(type, value, traceback)
                else:
                    raise value
            except StopIteration as exc:
                # Suppress StopIteration *unless* it's the same exception that
                # was passed to throw().  This prevents a StopIteration
                # raised inside the "with" statement from being suppressed.
                return exc is not value
            except RuntimeError as exc:
                # Don't re-raise the passed in exception. (issue27122)
                if exc is value:
                    return False
                # Likewise, avoid suppressing if a StopIteration exception
                # was passed to throw() and later wrapped into a RuntimeError
                # (see PEP 479).
                if type is StopIteration and exc.__cause__ is value:
                    return False
                raise
            except:
                # only re-raise if it's *not* the exception that was
                # passed to throw(), because __exit__() must not raise
                # an exception unless __exit__() itself failed.  But throw()
                # has to raise the exception to signal propagation, so this
                # fixes the impedance mismatch between the throw() protocol
                # and the __exit__() protocol.
                #
                if sys.exc_info()[1] is value:
                    return False
                raise
            raise RuntimeError("generator didn't stop after throw()")

    async def __aenter__(self):
        try:
            if inspect.isgenerator(self.gen):
                return self.gen.send(None)
            else:
                return await self.gen.asend(None)
        except (StopAsyncIteration, StopIteration):
            raise RuntimeError("generator didn't yield") from None

    async def __aexit__(self, type, value, traceback):
        if type is None:
            try:
                if inspect.isgenerator(self.gen):
                    self.gen.send(None)
                else:
                    await self.gen.asend(None)
            except (StopAsyncIteration, StopIteration):
                return False
            else:
                raise RuntimeError("generator didn't stop")
        else:
            if value is None:
                value = type()
            try:
                if inspect.isgenerator(self.gen):
                    await self.gen.throw(type, value, traceback)
                else:
                    await self.gen.athrow(type, value, traceback)
            except (StopAsyncIteration, StopIteration) as exc:
                return exc is not value
            except RuntimeError as exc:
                if exc is value:
                    return False
                if issubclass(type, (StopAsyncIteration, StopIteration))\
                        and exc.__cause__ is value:
                    return False
                raise
            except:
                if sys.exc_info()[1] is value:
                    return False
                raise
            raise RuntimeError("generator didn't stop after throw()")


class ConnectionManager(object):

    @staticmethod
    def debug_callback():
        return os.getenv("UNIT_TEST_MODE", "").lower() == "true"

    def __init__(self, debug_callback=None):
        if debug_callback:
            self.debug_callback = debug_callback

    def __call__(self, *args, debug_callback=None):
        """
        返回连接管理下的方法
        :param func:
        :return:
        """
        if debug_callback:
            return self.__class__(debug_callback)

        func = args[0]

        if getattr(func, "conn_ignore", False):
            return func

        if asyncio.iscoroutinefunction(func):
            @wraps(func)
            async def inner(self_or_cls, *args, **kwargs):
                if self.debug_callback():
                    return await func(self_or_cls, *args, **kwargs)

                callargs = get_callargs(func, self_or_cls, *args, **kwargs)
                callargs.pop("cls", None)
                gen = self_or_cls.get_store(self_or_cls, **callargs)
                async with gen as proxy_instance:
                    return await func(proxy_instance, *args, **kwargs)
        else:
            @wraps(func)
            def inner(self_or_cls, *args, **kwargs):
                if self.debug_callback():
                    return func(self_or_cls, *args, **kwargs)

                callargs = get_callargs(func, self_or_cls, *args, **kwargs)
                callargs.pop("cls", None)
                gen = self_or_cls.get_store(self_or_cls, **callargs)
                with gen as proxy_instance:
                    if proxy_instance is None:
                        warnings.warn("All DriverMixin lost efficacy，because "
                                      "of async mixin used with sync method.")
                        proxy_instance = self_or_cls
                    return func(proxy_instance, *args, **kwargs)
        return inner


conn_manager = ConnectionManager()


def conn_ignore(func):
    """
    使用了持久化元类时，使方法忽略使用连接管理，该装饰器必须紧靠方法
    :param func:
    :return:
    """
    func.conn_ignore = True
    return func


class DriverMixin(object):
    """
    配合conn_manager用来控制数据库访问。
    DriverMixin实现需要注意以下三点：
    1 所有Mixin都继承于DriverMixin(或其子类)，使用super调用父类的get_store方法，
      get_store需要被contextmanager装饰， contextmanager(非内置)来自于apistellar。
    2 Mixin支持将get_store实现成异步的，但最好是叶子节点类，
      异步Mixin最好不要被其它Mixin继承，除非你可以理清mro顺序。
    3 如果继承了异步Mixin创建业务类，那么其中的同步方法不能被conn_manager装饰。
      若装饰了，也不会有任何Mixin效果，还会收到警告。
    """

    @classmethod
    @contextmanager
    def get_store(cls, self_or_cls, **callargs):
        """
        子类需要通过super调用父类的get_store方法
        :param self_or_cls: 调用类方法和实例方法时的cls或self
        :param callargs: 方法调用时参数表
        :return: 返回proxy对象
        """
        yield self_or_cls


class PersistentMeta(type):
    """
    为实例方法和类方法增加conn_manager装饰器
    """
    def __new__(mcs, name, bases, attrs):
        for attr_name in attrs.keys():
            func = attrs[attr_name]
            # 去掉魔术方法和私有方法
            if isinstance(func, FunctionType) and not func.__name__.startswith(
                    "__"):
                attrs[attr_name] = conn_manager(func)
            if isinstance(func, classmethod):
                func = func.__func__
                if not func.__name__.startswith("__"):
                    attrs[attr_name] = classmethod(conn_manager(func))

        return super(PersistentMeta, mcs).__new__(mcs, name, bases, attrs)



import re
import typing

from enum import Enum
from urllib.parse import unquote
from collections import namedtuple
from flask.sessions import SecureCookieSession

from toolkit import cache_classproperty
from toolkit.settings import SettingsLoader, Settings

from .exceptions import Readonly

Cookie = typing.NewType('Cookie', str)
FormParam = typing.NewType("FormParam", str)
Session = typing.NewType("Session", SecureCookieSession)
MultiPartForm = typing.NewType("MultiPartForm", dict)
UrlEncodeForm = typing.NewType("UrlEncodeForm", dict)
DummyFlaskApp = namedtuple(
    "DummyFlaskApp",
    "session_cookie_name,secret_key,permanent_session_lifetime,config")


class File(object):
    mime_type_regex = re.compile(rb"Content-Type: (\S*)")
    content_length_regex = re.compile(rb"Content-Length: (\d*)")
    disposition_regex = re.compile(rb'Content-Disposition: form-data(?:; name\*?=\"?(?:(?P<name_enc>[\w\-]+?)\'(?P<name_lang>\w*)\')?(?P<name>[^\";]*)\"?)?.*?(?:; filename\*?=\"?(?:(?P<enc>[\w\-]+?)\'(?P<lang>\w*)\')?(?P<filename>[^\"]*?)\"?)?(?:$|\r\n)')

    def __init__(self, stream, receive, boundary, name,
                 filename, mimetype, content_length):
        self.content_length = content_length
        self.mimetype = mimetype
        self.receive = receive
        self.filename = filename
        self.name = name
        self.stream = stream
        self.tmpboundary = b"\r\n--" + boundary
        self.boundary_len = len(self.tmpboundary)
        self._last = b""
        self._size = 0
        self.body_iter = self._iter_content()

    def __aiter__(self):
        return self.body_iter

    def __str__(self):
        return f"<{self.__class__.__name__} " \
               f"name={self.name} " \
               f"filename={self.filename} >"

    __repr__ = __str__

    def iter_content(self):
        return self.body_iter

    async def _iter_content(self):
        stream = self.stream
        while True:
            # 如果存在read过程中剩下的，则直接返回
            if self._last:
                yield self._last
                continue

            index = self.stream.body.find(self.tmpboundary)
            if index != -1:
                # 找到分隔线，返回分隔线前的数据
                # 并将分隔及分隔线后的数据返回给stream
                read, stream.body = stream.body[:index], stream.body[index:]
                self._size += len(read)
                yield read
                if self._last:
                    yield self._last
                break
            else:
                if self.stream.closed:
                    raise RuntimeError("Uncomplete content!")
                # 若没有找到分隔线，为了防止分隔线被读取了一半
                # 选择只返回少于分隔线长度的部分body
                read = stream.body[:-self.boundary_len]
                stream.body = stream.body[-self.boundary_len:]
                self._size += len(read)
                yield read
                await self.get_message(self.receive, stream)

    async def read(self, size=10240):
        read = b""
        assert size > 0, (999, "Read size must > 0")
        while len(read) < size:
            try:
                buffer = await self.body_iter.asend(None)
            except StopAsyncIteration:
                return read
            read = read + buffer
            read, self._last = read[:size], read[size:]
        return read

    @staticmethod
    async def get_message(receive, stream):
        message = await receive()

        if not message['type'] == 'http.request':
            raise RuntimeError(
                f"Unexpected ASGI message type: {message['type']}.")

        if not message.get('more_body', False):
            stream.closed = True
        stream.body += message.get("body", b"")

    def tell(self):
        return self._size

    @classmethod
    async def from_boundary(cls, stream, receive, boundary):
        tmp_boundary = b"--" + boundary
        while not stream.closed:
            await cls.get_message(receive, stream)

            if b"\r\n\r\n" in stream.body and tmp_boundary in stream.body or \
                    stream.closed:
                break

        return cls(
            stream, receive, boundary, *cls.get_headers(stream, tmp_boundary))

    @classmethod
    def get_headers(cls, stream, tmp_boundary):
        end_boundary = tmp_boundary + b"--"
        body = stream.body
        index = body.find(tmp_boundary)
        if index == body.find(end_boundary):
            raise StopAsyncIteration
        body = body[index + len(tmp_boundary):]
        split_index = body.find(b"\r\n\r\n")
        header_str, body = body[:split_index], body[split_index + 4:]
        headers = cls.parse(header_str)
        stream.body = body
        return headers

    @classmethod
    def parse(cls, header_str):
        groups = cls.disposition_regex.search(header_str).groupdict()

        filename = groups["filename"] and unquote(groups["filename"].decode())
        if groups["enc"]:
            filename = filename.encode().decode(groups["enc"].decode())

        name = groups["name"] and unquote(groups["name"].decode())
        if groups["name_enc"]:
            name = name.encode().decode(groups["name_enc"].decode())

        mth = cls.mime_type_regex.search(header_str)
        mimetype = mth and mth.group(1).decode()
        mth = cls.content_length_regex.search(header_str) or 0
        content_length = int(mth and mth.group(1))
        assert name, "FileStream iterated without File consumed. "
        return name, filename, mimetype, content_length


class FileStream(object):

    def __init__(self, receive, boundary):
        self.receive = receive
        self.boundary = boundary
        self.body = b""
        self.closed = False

    def __aiter__(self):
        return self

    async def __anext__(self):
        return await File.from_boundary(self, self.receive, self.boundary)


class InheritType(Enum):
    DUPLICATE = 0  # 重名且类型相符，在参数列表和赋值中不体现，在super中体现
    OVERWRITE = 1  # 重名但是类型不同，全部位置要体现，但是需要父类的参数改名字
    NORMAL = 2  # 正常的情况


class Inject(object):

    def __init__(self, type, default):
        self.annotation = self.type = type
        self.name = type.__name__.lower()
        self.default = default

    def __set__(self, instance, value):
        raise Readonly(f"Readonly object of {self.type.__name__}.")

    def __get__(self, instance, cls):
        if instance:
            return instance.__dict__[self.name]
        else:
            return self


class InjectManager(object):

    def __init__(self, default=None):
        self.default = default

    def __lshift__(self, other):
        return Inject(other, self.default)

    def __call__(self, *, default=None):
        return InjectManager(default)


inject = InjectManager()


class SettingsMixin(object):
    settings_path = None

    def property(func):
        """
        这是一个假的property，它存在的意义在于误导pycharm，
        使cache_classproperty可以像property一样被注释类型。
        :param func:
        :return:
        """
        return func

    @property
    @cache_classproperty
    def settings(cls):
        # type: () -> Settings
        return SettingsLoader().load(SettingsMixin.settings_path or "settings")

    @classmethod
    def register_path(cls, settings_path):
        cls.settings_path = settings_path


del SettingsMixin.property



import os
import sys

from argparse import ArgumentParser
from os.path import join, abspath, dirname, exists
from jinja2 import Environment, FileSystemLoader

from ..helper import ArgparseHelper
from .tasks import *


class Command(object):
    """
    项目构建工具
    """

    def __init__(self, tasks):
        self.tasks = tasks
        self.args = self.parse_args()
        self.templates = [self.args.templates,
                          join(abspath(dirname(dirname(__file__))),
                               "templates")]
        self.task = tasks[self.args.task.lower()]()

    def create(self):
        self.task.template = self.make_sure_templates()
        env = Environment(loader=FileSystemLoader(self.templates))
        try:
            self.task.create(env, **vars(self.args))
        except AssertionError as e:
            print(e.args[0])

    def make_sure_templates(self):
        for template in self.templates[:]:
            if not template:
                self.templates.remove(template)
            else:
                if exists(join(template, self.args.task.lower())):
                    return template

    def parse_args(self):
        base_parser = ArgumentParser(
            description=self.__class__.__doc__,
            add_help=False,
            conflict_handler="resolve")
        base_parser.add_argument("-t", "--templates", help="模板路径.")

        parser = ArgumentParser(description="Apistar项目构建工具", add_help=False)
        parser.add_argument('-h', '--help', action=ArgparseHelper,
                            help='显示帮助信息并退出. ')
        sub_parsers = parser.add_subparsers(dest="task", help="创建模块类型. ")

        for name, task in self.tasks.items():
            sub_parser = sub_parsers.add_parser(
                name.lower(), parents=[base_parser],
                help=task.__doc__, conflict_handler="resolve")
            task.enrich_parser(sub_parser)
        # 当不提供参数时，len(sys.argv) == 1, 不会打印帮助，需要手动打印
        # add_subparsers无法指定required=True, 这是argparse的bug。
        if len(sys.argv) < 2:
            parser.print_help()
            exit(1)
        return parser.parse_args()


def find_tasks():
    sys.path.insert(0, os.getcwd())
    tasks = dict()

    try:
        model = __import__("tasks")
        for k in dir(model):
            v = getattr(model, k)
            if v is not Task and isinstance(v, type) and issubclass(v, Task):
                tasks[k.lower()] = v
    except ImportError:
        pass

    for k, v in globals().items():
        if v is not Task and isinstance(v, type) and issubclass(v, Task):
            tasks[k.lower()] = v
    return tasks


def main():
    Command(find_tasks()).create()


if __name__ == "__main__":
    main()



import os
import re
import sys
import json
import time
import glob
import email
import types
import inspect
import logging

from urllib.parse import urljoin
from aiohttp import ClientSession
from functools import wraps, reduce
from collections.abc import Mapping
from pyaop import Proxy, Return, AOP
from types import FunctionType, MethodType
from asyncio import Future, get_event_loop
from argparse import Action, _SubParsersAction

from apistar import Include, Route
from apistar.http import PathParams, Response
from apistar.server.asgi import ASGIReceive, ASGIScope, ASGISend

from werkzeug._compat import string_types
from werkzeug.utils import escape, text_type
from werkzeug.http import dump_cookie, dump_header, parse_set_header


def get_real_method(obj, name):
    """
    IPython bug fix.
    """
    try:
        canary = getattr(obj, '_ipython_canary_method_should_not_exist_', None)
        if isinstance(canary, Future):
            canary.cancel()
    except Exception:
        return None

    if canary is not None:
        # It claimed to have an attribute it should never have
        return None

    try:
        m = getattr(obj, name, None)
        if isinstance(canary, Future):
            canary.cancel()
    except Exception:
        return None

    if inspect.isclass(obj) and not isinstance(m, types.MethodType):
        return None

    if callable(m):
        return m

    return None


def enhance_response(resp):

    def set_cookie(self, key, value='', max_age=None, expires=None,
                   path='/', domain=None, secure=False, httponly=False,
                   samesite=None):
        """Sets a cookie. The parameters are the same as in the cookie `Morsel`
        object in the Python standard library but it accepts unicode data, too.

        A warning is raised if the size of the cookie header exceeds
        :attr:`max_cookie_size`, but the header will still be set.

        :param key: the key (name) of the cookie to be set.
        :param value: the value of the cookie.
        :param max_age: should be a number of seconds, or `None` (default) if
                        the cookie should last only as long as the client's
                        browser session.
        :param expires: should be a `datetime` object or UNIX timestamp.
        :param path: limits the cookie to a given path, per default it will
                     span the whole domain.
        :param domain: if you want to set a cross-domain cookie.  For example,
                       ``domain=".example.com"`` will set a cookie that is
                       readable by the domain ``www.example.com``,
                       ``foo.example.com`` etc.  Otherwise, a cookie will only
                       be readable by the domain that set it.
        :param secure: If `True`, the cookie will only be available via HTTPS
        :param httponly: disallow JavaScript to access the cookie.  This is an
                         extension to the cookie standard and probably not
                         supported by all browsers.
        :param samesite: Limits the scope of the cookie such that it will only
                         be attached to requests if those requests are
                         "same-site".
        """
        self.headers['Set-Cookie'] = dump_cookie(
            key,
            value=value,
            max_age=max_age,
            expires=expires,
            path=path,
            domain=domain,
            secure=secure,
            httponly=httponly,
            charset=self.charset,
            max_size=self.max_cookie_size,
            samesite=samesite
        )

    def delete_cookie(self, key, path='/', domain=None):
        """Delete a cookie.  Fails silently if key doesn't exist.

        :param key: the key (name) of the cookie to be deleted.
        :param path: if the cookie that should be deleted was limited to a
                     path, the path has to be defined here.
        :param domain: if the cookie that should be deleted was limited to a
                       domain, that domain has to be defined here.
        """
        self.set_cookie(key, expires=0, max_age=0, path=path, domain=domain)

    setattr(resp, "set_cookie", set_cookie)
    setattr(resp, "delete_cookie", delete_cookie)
    setattr(resp, "max_cookie_size", 4093)

    def _set_property(name, doc=None):
        def fget(self):
            def on_update(header_set):
                if not header_set and name in self.headers:
                    del self.headers[name]
                elif header_set:
                    self.headers[name] = header_set.to_header()

            return parse_set_header(self.headers.get(name), on_update)

        def fset(self, value):
            if not value:
                del self.headers[name]
            elif isinstance(value, string_types):
                self.headers[name] = value
            else:
                self.headers[name] = dump_header(value)

        return property(fget, fset, doc=doc)

    setattr(resp, "vary", _set_property(
        "vary", doc='''
         The Vary field value indicates the set of request-header fields that
         fully determines, while the response is fresh, whether a cache is
         permitted to use the response to reply to a subsequent request
         without revalidation.'''))


def find_children(cls, initialize=True, kwargs=None):
    """
    获取所有(component)的子类或其实例。
    :param cls: 父类
    :param initialize: 是否生成实例
    :param kwargs: 若初始化，关键字参数
    :return:
    """
    def _load(cs):
        children = []
        for c in cs:
            children.append(c)
            children.extend(_load(c.__subclasses__()))
        return children

    if initialize:
        kwargs = kwargs or dict()
    return [c(**kwargs) if initialize else c for c in _load(cls.__subclasses__())]


def walk_packages(current_path):
    """
    加载当前路径下的所有package，使得其中的Controller子类得以激活
    加载一个包时，如果包下面有子包，只需要导入子包，父包也会一起
    被加载。项目约定Controller子类必须定义在包中(__init__.py)。
    所以只考虑加载所有包，不考虑加载其它模块。
    :param current_path:
    :return:
    """
    files = glob.glob(os.path.join(current_path, "*"))
    find_dir = False

    for file in files:
        if os.path.isdir(file):
            find_dir = True
            if not walk_packages(file):
                __import__(file.replace("/", ".").strip("."))

    return find_dir


def walk_modules(current_path, app_name=None):
    """
    由于walk_packages只加载到包。所以在只能将Controller定义在包中，
    但在引用service层时，各个包中的service层相互引用，有可能会出现循环引用，
    所以还是遍历所有模块吧，但是启动时的性能会差一些。
    :param current_path:
    :return:
    """
    current_path = current_path.rstrip("/")
    for root, _, filenames in os.walk(os.path.join(current_path, app_name, app_name)):
        imported = False
        for fn in filenames:
            if fn.endswith(".py") and not fn.startswith("_"):
                fn = fn[:-3]
                module_name = os.path.join(root, fn).replace(current_path, "")\
                    .replace("/", ".").strip(".")
                __import__(module_name)
                imported = True
        # 如果整个包都没有被导入过，则导入一下这个包，主要是为了加载__init__.py
        if not imported:
            __import__(root.replace(current_path, "").replace("/", ".").strip("."))


load_packages = walk_modules


def routing(controller):
    """
    获取当前Controller下所有route及其子Controller组成的Include
    :param controller:
    :return:
    """
    if not hasattr(controller, "prefix"):
        raise RuntimeError(f"{controller} is not routed! ")
    routes = []
    instance = controller()
    for name in vars(controller).keys():
        prop = getattr(controller, name)
        if hasattr(prop, "routes"):
            for route in prop.routes:
                route.controller = instance
                add_annotation(route.handler, _find_ancestor(controller))
                routes.append(route)

    for child_controller in controller.__subclasses__():
        child_include = routing(child_controller)
        if child_include:
            routes.append(child_include)

    if routes:
        return Include(controller.prefix, controller.name, routes)


def add_annotation(method, annotation, arg_name="self"):
    """
    如果定义的方法存在arg_name， 则为其加上类型注释.
    :param method:
    :param annotation:
    :param arg_name:
    :return:
    """
    if hasattr(method, "__code__") and \
            arg_name in method.__code__.co_varnames and \
            arg_name not in method.__annotations__:
        method.__annotations__[arg_name] = annotation


class ArgparseHelper(Action):
    """
        显示格式友好的帮助信息
    """

    def __init__(self,
                 option_strings,
                 dest="",
                 default="",
                 help=None):
        super(ArgparseHelper, self).__init__(
            option_strings=option_strings,
            dest=dest,
            default=default,
            nargs=0,
            help=help)

    def __call__(self, parser, namespace, values, option_string=None):
        parser.print_help()
        subparsers_actions = [
            action for action in parser._actions
            if isinstance(action, _SubParsersAction)]
        for subparsers_action in subparsers_actions:
            for choice, subparser in subparsers_action.choices.items():
                print("Command '{}'".format(choice))
                print(subparser.format_usage())

        parser.exit()


def parse_date(ims):
    """
    Parse rfc1123, rfc850 and asctime timestamps and return UTC epoch.
    """
    try:
        ts = email.utils.parsedate_tz(ims)
        return time.mktime(ts[:8] + (0,)) - (ts[9] or 0) - time.timezone
    except (TypeError, ValueError, IndexError, OverflowError):
        return None


def parse_range_header(header, maxlen=0):
    """
    Yield (start, end) ranges parsed from a HTTP Range header. Skip
        unsatisfiable ranges. The end index is non-inclusive.
    :param header:
    :param maxlen:
    :return:
    """
    if not header or header[:6] != 'bytes=': return
    ranges = [r.split('-', 1) for r in header[6:].split(',') if '-' in r]
    for start, end in ranges:
        try:
            if not start:  # bytes=-100    -> last 100 bytes
                start, end = max(0, maxlen-int(end)), maxlen
            elif not end:  # bytes=100-    -> all but the first 99 bytes
                start, end = int(start), maxlen
            else:          # bytes=100-200 -> bytes 100-200 (inclusive)
                start, end = int(start), min(int(end)+1, maxlen)
            if 0 <= start < end <= maxlen:
                yield start, end
        except ValueError:
            pass


async def file_iter_range(fp, offset, bytes, maxread=1024*1024):
    """
    Yield chunks from a range in a file. No chunk is bigger than maxread.
    :param fp:
    :param offset:
    :param bytes:
    :param maxread:
    :return:
    """
    await fp.seek(offset)
    while bytes > 0:
        part = await fp.read(min(bytes, maxread))
        if not part: break
        bytes -= len(part)
        yield part


class TypeEncoder(json.JSONEncoder):
    options = {Mapping: dict}

    @classmethod
    def register(cls, type_mapping):
        cls.options.update(type_mapping)

    def default(self, obj):
        for key, val in self.options.items():
            if isinstance(obj, key):
                return val(obj)
        return json.JSONEncoder.default(self, obj)


class HookReturn(Exception):
    pass


def redirect(location, code=302, Response=None):
    """Returns a response object (a ASGI application) that, if called,
    redirects the client to the target location.  Supported codes are 301,
    302, 303, 305, and 307.  300 is not supported because it's not a real
    redirect and 304 because it's the answer for a request with a request
    with defined If-Modified-Since headers.

    .. versionadded:: 0.6
       The location can now be a unicode string that is encoded using
       the :func:`iri_to_uri` function.

    .. versionadded:: 0.10
        The class used for the Response object can now be passed in.

    :param location: the location the response should redirect to.
    :param code: the redirect status code. defaults to 302.
    :param class Response: a Response class to use when instantiating a
        response. The default is :class:`werkzeug.wrappers.Response` if
        unspecified.
    """
    if Response is None:
        from apistar.http import HTMLResponse as Response

    display_location = escape(location)
    if isinstance(location, text_type):
        # Safe conversion is necessary here as we might redirect
        # to a broken URI scheme (for instance itms-services).
        from werkzeug.urls import iri_to_uri
        location = iri_to_uri(location, safe_conversion=True)
    response = Response(
        '<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 3.2 Final//EN">\n'
        '<title>Redirecting...</title>\n'
        '<h1>Redirecting...</h1>\n'
        '<p>You should be redirected automatically to target URL: '
        '<a href="%s">%s</a>.  If not click the link.' %
        (escape(location), display_location),
        headers={"Location": location},
        status_code=code)
    return response


class ChildrenFactory(object):

    def __init__(self, father, kwargs=None):
        self.father = father
        self.kwargs = kwargs or {}
        self.found = dict()

    def install(self, **kwargs):
        self.found.clear()
        self.kwargs.update(kwargs)

    def _get_mapping(self):
        return {getattr(child, "name", child.__name__.lower()): child
                for child in find_children(self.father, False)}

    def __contains__(self, item):
        result = item in self.found
        if result is False:
            result = item in self._get_mapping()
        return result

    def get(self, item, default=None):
        try:
            return self[item]
        except KeyError:
            return default

    def __getitem__(self, item):
        child = self.found.get(item)
        if child is None:
            mapping = self._get_mapping()
            child = mapping[item](**self.kwargs)
            self.found[item] = child
        return child

    def __setitem__(self, item, value):
        self.found[item] = value


def require(
        container_cls,
        judge=lambda container: container.user,
        error="Login required!"):
    """
    装饰一个可被注入的函数，注入container_cls的实例，
    并调用judge判断其是否符合条件，否则抛出异常，异常信息为error。
    注：在controller handler中使用时，必须装饰在路由装饰器前。
    :param container_cls:
    :param judge:
    :param error:
    :return:
    """
    def auth(func):
        args = inspect.getfullargspec(func).args
        args_def = ", ".join(args)
        func_def = """
@wraps(func)
async def wrapper(__container, {}):
    from collections.abc import Awaitable
    assert judge(__container), (401, "{}")
    awaitable = func({})
    if isinstance(awaitable, Awaitable):
        return await awaitable
    return awaitable
        """.format(args_def, error, args_def)
        return _build_new_func(
            func_def, func, {"judge": judge}, {"__container": container_cls})

    return auth


def _build_new_func(func_def, func, nps=None, ans=None):
    namespace = dict(__name__='entries_%s' % func.__name__)
    if nps:
        namespace.update(nps)

    namespace["func"] = func
    namespace["wraps"] = wraps
    exec(func_def, namespace)
    wrapper = namespace["wrapper"]
    new_func = types.FunctionType(
        wrapper.__code__,
        wrapper.__globals__,
        wrapper.__name__,
        func.__defaults__)
    new_func.__annotations__.update(func.__annotations__)
    # 增加返回值封装信息
    if hasattr(func, "__return_wrapped"):
        new_func.__return_wrapped = func.__return_wrapped
    new_func.__doc__ = func.__doc__
    if ans:
        new_func.__annotations__.update(ans)
    return new_func


def return_wrapped(success_code=0, success_key_name="data", error_info=None):
    """
    为handler的返回值提供默认的成功返回码及返回信息对应的key名称，必须装饰在路由装饰器前。
    :param success_code:
    :param success_key_name:
    :param error_info:
    :return:
    """

    assert isinstance(success_key_name, str), "`success_key_name` must be str!"
    assert isinstance(success_code, int), "`success_code` must be int!"

    def return_wrapper(func):
        args = inspect.getfullargspec(func).args
        args_def = ", ".join(args)
        func_def = """
@wraps(func)
async def wrapper({}):
    from collections.abc import Awaitable
    awaitable = func({})
    if isinstance(awaitable, Awaitable):
        awaitable = await awaitable
    return_val = dict(code=success_code)
    return_val[success_key_name] = awaitable
    return return_val
        """.format(args_def, args_def)
        nps = {"success_key_name": success_key_name,
               "success_code": success_code,
               "error_info": error_info}
        new_func = _build_new_func(func_def, func, nps)
        new_func.__return_wrapped = nps
        return new_func
    return return_wrapper


class MySelf(object):
    def __getattr__(self, item):
        return self

    def __getitem__(self, item):
        if item == 'type':
            return "http.request"
        if item == "server":
            return "", 80
        return self

    def __call__(self, *args, **kwargs):
        return self

    def __bool__(self):
        return False

    def __str__(self):
        return ""

    def __radd__(self, obj):
        return obj + type(obj)(self)

    def __iter__(self):
        return iter([])

    def __await__(self):
        loop = get_event_loop()
        future = loop.create_future()
        future.set_result(self)
        return future.__await__()

    __repr__ = __str__


async def add_success_callback(fut, callback):
    """
    这个方法的作用相当于future.set_done_callback。
    :param fut:
    :param callback:
    :return:
    """
    try:
        result = await fut
    except Exception as e:
        result = e
    callback(result)
    return result


class NotImplementedProp(object):
    """
    用来对子类需要实现的类属性进行占位
    """
    def __get__(self, instance, owner):
        return NotImplemented


class classproperty(object):
    """
    property只能用于实例方法到实例属性的转换，使用classproperty来支持类方法到类属性的转换
    """
    def __init__(self, func):
        self.func = func

    def __get__(self, instance, owner):
        return self.func(owner)


def cache_classproperty(func):
    """
    缓存类属性，只计算一次
    :param func:
    :return:
    """
    @classproperty
    @wraps(func)
    def wrapper(*args, **kwargs):
        prop_name = "_" + func.__name__
        if prop_name not in args[0].__dict__:
            setattr(args[0], prop_name, func(*args, **kwargs))
        return args[0].__dict__[prop_name]
    return wrapper


def proxy(obj, prop, prop_name):
    """
    为object对象代理一个属性
    :param obj:
    :param prop: 属性
    :param prop_name: 属性名
    :return:
    """
    assert isinstance(prop_name, str), "prop_name must be string!"

    def common(proxy, name, value=None):
        if name == prop_name:
            if value:
                raise RuntimeError(f"{prop_name} readonly!")
            else:
                Return(prop)

    return Proxy(obj, before=[
        AOP.Hook(common, ["__getattribute__", "__setattr__", "__delattr__"]),
        ])


class RestfulApi(object):

    @cache_classproperty
    def logger(cls):
        """
        增加cache_classproperty意义在于懒加载logger,
        以避免format过早生效导致一些基础日志打印失败
        :return:
        """
        logger = logging.getLogger(cls.__name__.lower())
        logger.addHandler(logging.StreamHandler(sys.stdout))
        return logger

    def __init__(self, host, port):
        super().__init__()
        self.host = host
        self.port = port
        self.prefix = f'http://{host}:{port}'

    def url(self, path):
        return urljoin(self.prefix, path)


def path_repl(mth):
    return "{" + mth.group(1).lstrip("+") + "}"


def _find_ancestor(cls, until_not_have=None):
    """
    查找祖先，直到祖先为object或者祖先没有until_not_have这个属性时，返回当前类。
    :param cls:
    :return:
    """
    if until_not_have and not hasattr(cls.__base__, until_not_have) \
            or cls.__base__ == object:
        return cls
    return _find_ancestor(cls.__base__)


def get_callargs(func, *args, **kwargs):
    """
    找到层层装饰器下最里层的函数的callargs
    :param func:
    :param args:
    :param kwargs:
    :return:
    """
    for closure in func.__closure__ or []:
        if isinstance(closure.cell_contents, (FunctionType, MethodType)):
            func = closure.cell_contents
            return get_callargs(func, *args, **kwargs)

    args = inspect.getcallargs(func, *args, **kwargs)
    spec = inspect.getfullargspec(func)
    if spec.varkw:
        args.update(args.pop(spec.varkw, {}))

    return args


def register(url, path=None, error_check=None, conn_timeout=9,
             read_timeout=9, have_path_param=False):
    """
    为RPC方法注册路由
    :param url:
    :param path: 在返回的响应数据中，如果是json格式，有用数据位置如`data.value
    `对应{"data": {"value": "有用的数据"}}，非json不需要填写。
    :param error_check: 在返回的响应数据中，如果是json格式，
    用来检查该json是否有效的回调函数，一般会配合响应码来检查。
    :param conn_timeout:
    :param read_timeout:
    :param have_path_param: 是否有restful风格的路径参数
    :return:
    """
    def request_wrapper(func):
        @wraps(func)
        async def inner(*args, **kwargs):
            cookies = kwargs.pop("cookies", None)

            async with ClientSession(
                    conn_timeout=conn_timeout, read_timeout=read_timeout,
                    cookies=cookies) as session:
                self = args[0]
                u = _find_ancestor(self.__class__, "url").url(self, url)

                if have_path_param:
                    callargs = get_callargs(func, *args, **kwargs)
                    path_params = callargs.pop("path_params", None)
                    u = re.sub('{([^}]*)}', path_repl, u).format(**path_params)
                self.logger.debug("Search url: %s" % u)
                self.logger.debug(
                    "Search query, args: %s, kwargs %s. " % (args[1:], kwargs))
                self = proxy(proxy(self, u, "url"), session, "session")
                data = None

                try:
                    data = await func(self, *args[1:], **kwargs)
                    if error_check and isinstance(
                            data, dict) and error_check(data):
                        raise BackendServiceError(data)
                except BackendServiceError as e:
                    raise e
                except Exception as e:
                    self.logger.error(
                        f"Error in calling {self.url}.return: {data}")
                    raise BackendInternalError() from e
                return path_parse(path, data)

        return inner
    return request_wrapper


def _val_get(data, y):
    try:
        return data[y]
    except TypeError as e:
        if y.isdigit():
            return data[int(y)]
        else:
            raise e


def path_parse(path, data):
    """
    从字典中获取指定路径下的数据
    :param path: "a.b.1.c"
    :param data: {"a": {"b": [{"c": 3}, {"c": 4}]}}
    :return: 4
    """
    if not path:
        return data
    return reduce(_val_get, path.split("."), data)


class BackendInternalError(RuntimeError):
    pass


class BackendServiceError(RuntimeError):
    pass


# mock state
STATE = {
        'scope': MySelf(),
        'receive': MySelf(),
        'send': MySelf(),
        'exc': None,
        'path_params': MySelf(),
        'route': MySelf()
        }


INITIAL = {
            'scope': ASGIScope,
            'receive': ASGIReceive,
            'send': ASGISend,
            'exc': Exception,
            'path_params': PathParams,
            'route': Route,
            'response': Response,
        }



import json
import asyncio
import logging

from abc import ABCMeta
from apistar import http
from functools import partial
from _collections_abc import _check_methods

from ..helper import TypeEncoder


logger = logging.getLogger("websocket")


class WebSocketHandler(metaclass=ABCMeta):

    async def websocket_connect(self, message, path: http.Path):
        logger.debug(f"Websocket of {path} connect. ")
        return ""

    async def websocket_disconnect(self, message, path: http.Path):
        logger.debug(f"Websocket of {path} disconnect. ")

    @classmethod
    def __subclasshook__(cls, C):
        if cls is WebSocketHandler:
            try:
                return _check_methods(
                    C, "websocket_connect", "websocket_disconnect", "websocket_receive")
            except AttributeError:
                return False
        return NotImplemented


class WebSocketApp:
    """
    每个websocket链接都是独立实例
    """
    persist = True
    handler = None
    state = None
    _send = None

    reverse_type = {
        "websocket.disconnect": "websocket.close",
        "websocket.receive": "websocket.send",
        "websocket.connect": "websocket.accept"
    }

    def __init__(self, scope, app):
        self.scope = scope
        self.app = app

    async def __call__(self, receive, send):
        state = {
            'scope': self.scope,
            'receive': receive,
            'send': send,
            'exc': None,
            'app': self,
            'path_params': None,
            'route': None
        }
        method = self.scope['method']
        path = self.scope['path']
        self._send = send
        try:
            route, path_params = self.app.router.lookup(path, method)
            state['route'] = route
            state['path_params'] = path_params
            self.state = state
            self.handler = route.handler

            if issubclass(self.handler, WebSocketHandler):
                self.handler = self.handler(self.send)
            # websocket建立之后持续处理收到的消息
            while self.persist:
                await self.handle(await receive())
        except Exception as exc:
            state["exc"] = exc
            await self.exception_handler(exc)

    async def exception_handler(self, exc: Exception):
        await self.send(str(exc), "websocket.close")
        raise exc

    async def send(self, buf, type="websocket.send"):
        message = {"type": type}

        if isinstance(buf, bytes):
            message["bytes"] = buf
        elif buf is None:
            message["type"] = "websocket.close"
        elif not isinstance(buf, str):
            buf = json.dumps(buf, cls=TypeEncoder)

        if isinstance(buf, str):
            message["text"] = buf
        await self._send(message)

    async def handle(self, message):
        """
        消息处理方法
        :param message:
        :return:
        """
        message_type = message["type"].replace(".", "_")
        # 如果handler是WebSocketHandler的子类实例或者duck type实例，
        # 则从中获取信息类型的处理方法。
        if isinstance(self.handler, WebSocketHandler):
            handler = getattr(self.handler, message_type)
        # 否则从WebSocketHandler获取处理方法，
        # WebSocketHandler未定义websocket.receive的处理器，
        # 因此直接使用handler来处理。
        else:
            handler = getattr(WebSocketHandler(), message_type, self.handler)

        wrapped = partial(handler, message)
        if asyncio.iscoroutinefunction(handler):
            wrapped._is_coroutine = asyncio.coroutines._is_coroutine
        if message_type == "websocket_disconnect":
            self.persist = False
        buf = await self.app.injector.run_async([wrapped], self.state)
        await self.send(buf, self.reverse_type[message["type"]])



import os
import sys

from apistar.server.asgi import ASGI_COMPONENTS
from apistar.server.injector import ASyncInjector
from apistar.server.validation import VALIDATION_COMPONENTS

from apistellar.bases.entities import SettingsMixin
from apistellar.helper import load_packages, find_children, STATE, INITIAL
from apistellar.bases.components import Component, ValidateRequestDataComponent


class Manager(object):

    @staticmethod
    def initialize(path, app_name=None):
        # 这一步是为了将settings.py所以目录加入可搜索路径
        sys.path.insert(0, path)
        path, app_name = os.path.split(path.rstrip("/"))
        sys.path.insert(0, path)
        load_packages(path, app_name)

    def finalize(self, settings="settings"):
        SettingsMixin.register_path(settings)
        INITIAL["app"] = self.__class__
        self.state = STATE
        self.state["app"] = self
        self.components = find_children(Component)
        self.components.append(ValidateRequestDataComponent())
        self.injector = ASyncInjector(
            list(ASGI_COMPONENTS + VALIDATION_COMPONENTS) + self.components,
            INITIAL)



# -*- coding:utf-8 -*-

from apistar import Route
from apistar.server import core
from apistar.codecs import jsonschema
from apistellar import types


__all__ = ["route", "get", "post", "delete", "put", "options"]

# mock掉apistar中的types和validators
jsonschema.validators = core.validators = types.validators
jsonschema.types = core.types = types


def route(prefix, name=None):
    """
    route为Controller指定名称和路径前缀
    :param prefix:
    :param name:
    :return:
    """
    def prefix_wrapper(cls):
        cls.prefix = "" if prefix == "/" else prefix
        cls.name = name or cls.__name__.lower()
        return cls

    return prefix_wrapper


def wrapper_method(method_name):
    """
    生成绑定http method的route方法
    @get()
    async def hello():
        return {"message": "hello world"}
    :param method_name: get
    :return:
    """
    def method(url=None, name=None, documented=True, standalone=False):

        def endpoint_wrapper(handler):
            # 重新声明的变量不能是url, 下同，否则声明提前会覆盖闭包的变量
            u = url or "/" + handler.__name__
            if u[0] != "/":
                u = "/" + u
            routes = getattr(handler, "routes", [])
            # 对于一个handler拥有多个方法装饰器的情况
            # 由于apistar不支持一个名字映射多个路由
            # 所以之后的装饰器生成的名字会使用_连接方法名
            if not routes:
                n = name or handler.__name__
            else:
                n = (name or handler.__name__) + "_" + method_name
            route = Route(
                u, method_name.upper(), handler, n, documented, standalone)

            if isinstance(handler, type):
                handler.routes = [route]
            else:
                handler.__dict__.setdefault("routes", []).append(route)

            return handler

        return endpoint_wrapper

    method.__name__ = method_name
    method.__qualname__ = method_name
    return method


get = wrapper_method("get")
post = wrapper_method("post")
delete = wrapper_method("delete")
options = wrapper_method("options")
put = wrapper_method("put")
websocket = get



import sys
import json
import typing
import logging
import traceback

from apistar import App, http
from flask.sessions import SecureCookieSessionInterface

from ..helper import HookReturn
from .entities import Session, DummyFlaskApp


class SessionHook(object):

    def __init__(self):
        self.session_interface = SecureCookieSessionInterface()

    def on_response(self,
                    app: DummyFlaskApp,
                    resp: http.Response,
                    session: Session):
        if session is not None:
            self.session_interface.save_session(app, session, resp)


class ErrorHook(object):
    """
    处理异常增加响应码
    """
    errors = {999: "Unknown error"}
    fmt = '[{asctime}] {name} {levelname}: {message}'

    def __init__(self):
        self.logger = logging.getLogger("ErrorHook")
        self.logger.propagate = False
        handler = logging.StreamHandler(sys.stdout)
        handler.setLevel(10)
        handler.setFormatter(logging.Formatter(self.fmt, style="{"))
        self.logger.addHandler(handler)

    def on_error(self, error: Exception, app: App) -> http.Response:
        """
        Handle error
        """
        args = list()
        if error.args:
            data = error.args[0]
            if not isinstance(data, (tuple, list)):
                if isinstance(data, int):
                    args.append(data)
                    args.append(None)
                else:
                    args.append(999)
                    args.append(data)
            else:
                args.extend(data)
        else:
            args.append(None)

        if not isinstance(args[0], int):
            args.insert(0, 999)

        args.extend([None, None, None])
        code, message, extra, *_ = args
        # apistar不支持在on_request时打断后续执行直接返回response
        # 所以在只能通过raise异常来通过异常参数传递响应。
        if isinstance(message, http.Response):
            return message

        if message is None:
            message = self.errors.get(code, "Not configured error")

        payload = {
            "type": "normal",
            "code": code,
            "errcode": getattr(message, "code", code),
            "message": message,
            "extra": extra,
        }
        detail = "".join(traceback.format_exc())
        if app.debug:
            payload["detail"] = detail
        self.logger.error(
            "Error happened: %s extra: %s trace:\n%s. ", message, extra, detail)
        return http.JSONResponse(payload)

    @classmethod
    def register(cls,
                 errors: typing.Union[typing.List[typing.Tuple[int, str]],
                                      typing.Mapping[int, str]]):
        cls.errors.update(errors)


class AccessLogHook(object):
    fmt = '{host} - - [{asctime}] {method} {path} {protocol}' \
          ' {status} {content_length} {agent}'

    def __init__(self):
        self.logger = logging.getLogger("access")
        self.logger.propagate = False
        handler = logging.StreamHandler(sys.stdout)
        handler.setLevel(10)
        handler.setFormatter(logging.Formatter(self.fmt, style="{"))
        self.logger.addHandler(handler)

    def on_response(self,
                    host: http.Host,
                    path: http.Path,
                    protocol: http.Scheme,
                    method: http.Method,
                    resp: http.Response,
                    user_agent: http.Header,
                    string: http.QueryString):
        if string:
            path = path + "?" + string
        self.log(host, path, protocol, method, resp.status_code,
                 resp.headers["Content-Length"], user_agent)
        return resp

    on_error = on_response

    def log(self, host, path, protocol, method, status, content_length, agent):
        self.logger.info("", extra={"host": host,
                                    "path": path,
                                    "protocol": protocol,
                                    "method": method,
                                    "status": status,
                                    "content_length": content_length,
                                    "agent": agent})


class Hook(object):
    """
    Hook基类，继承自此基类的hook可以自动发现。
    """
    order = 1


def Return(return_value):
    if isinstance(return_value, str):
        return_value = http.HTMLResponse(return_value)
    elif not isinstance(return_value, http.Response):
        return_value = http.JSONResponse(return_value)
    raise HookReturn(return_value)


# -*- coding:utf-8 -*-
import os
import sys

from abc import ABC, abstractmethod
from contextlib import contextmanager

from toolkit import cache_property, load

from apistellar.bases.controller import Controller
from apistellar.helper import load_packages, routing

from .parser import LogParser


class Painter(ABC):
    """
    用来处理routes并生成描述信息
    """
    routes = None

    def __init__(self, current_dir):
        self.current_dir = current_dir

    @cache_property
    def routes(self):
        """
        获取所有routes
        :return:
        """
        # 这一步是为了将settings.py所以目录加入可搜索路径
        sys.path.insert(0, self.current_dir)
        self.current_dir, app_name = os.path.split(self.current_dir.rstrip("/"))
        sys.path.insert(0, self.current_dir)

        load_packages(self.current_dir, app_name)
        include = routing(Controller)
        routes = [include] if include else []
        return routes

    @abstractmethod
    def paint(self):
        """
        遍历routes渲染描述信息
        :return:
        """


class LogPainter(Painter):
    """
    打印route的简单信息，如日志等
    """

    def __init__(self, write, format, current_dir):
        super(LogPainter, self).__init__(current_dir)
        self.write = write
        self.format = format
        self.parser = LogParser()

    @contextmanager
    def paint(self):
        yield self.routes

        if not self.routes:
            self.write("Noting to route. ")
            return

        for args in self.parser.parse_docs(self.routes):
            self.write(self.format(*args))


class AppLogPainter(LogPainter):
    """
    在app启动时的routes日志输出
    """

    def __init__(self, write, current_dir):
        super(AppLogPainter, self).__init__(write, self._format, current_dir)

    @staticmethod
    def _format(method, parttern, name, ca_name):
        return f"Route method: {method}, url: {parttern} to {name}."


class ShowLogPainter(LogPainter):
    """
    查看routes时使用
    """

    def __init__(self, format):
        super(ShowLogPainter, self).__init__(print, format, os.getcwd())


class DocPainter(Painter):
    """
    用来生成接口文档
    """

    def __init__(self, current_dir, parser_path):
        super(DocPainter, self).__init__(current_dir)
        self.parser = load(parser_path)()

    def paint(self):
        return self.parser.parse_docs(self.routes)



class Mocker(object):

    def __init__(self):
        self.datas = []

    def add_mock(self, type, value, param_name=None):
        self.datas.append((type, value, param_name))

    def __iter__(self):
        return iter(self.datas)
